/*****************************************************************************
  Copyright (c) 2000, Intel Corporation
 *****************************************************************************/

/* Macros to make drivers compatible with 2.2 - 2.4 Linux kernels
 *
 * In order to make a single network driver work with all 2.2-2.4 kernels
 * these compatibility macros can be used.
 * They are backwards compatible implementations of the latest APIs.
 * The idea is that these macros will let you use the newest driver with old
 * kernels, but can be removed when working with the latest and greatest.
 */

#ifndef E1000_LINUX_KERNEL_COMPAT_H
#define E1000_LINUX_KERNEL_COMPAT_H

#include <linux/version.h>

/*****************************************************************************
 **
 **  PCI Bus Changes
 **
 *****************************************************************************/

/* Accessing the BAR registers from the PCI device structure
 * Changed from base_address[bar] to resource[bar].start in 2.3.13
 * The pci_resource_start inline function was introduced in 2.3.43 
 */
#if   ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,13) )
#define pci_resource_start(dev, bar) (dev)->base_address[(bar)]
#elif ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43) )
#define pci_resource_start(dev, bar) (dev)->resource[(bar)].start
#endif

/* Starting with 2.3.23 drivers are supposed to call pci_enable_device
 * to make sure I/O and memory regions have been mapped and potentially 
 * bring the device out of a low power state
 */
#if   ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,23) )
#define pci_enable_device(dev) do{} while(0)
#endif

/*****************************************************************************
 **
 **  Network Device API Changes
 **
 *****************************************************************************/

/* In 2.3.14 the device structure was renamed to net_device 
 */
#if   ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,14) )
#define net_device device
#endif

/* 'Softnet' network stack changes merged in 2.3.43 
 * these are 2.2 compatible defines for the new network interface API
 * 2.3.47 added some more inline functions for softnet to remove explicit 
 * bit tests in drivers
 */
#if   ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43) )
#define netif_start_queue(dev)   clear_bit  (0, &(dev)->tbusy)
#define netif_stop_queue(dev)    set_bit    (0, &(dev)->tbusy)
#define netif_wake_queue(dev)    { clear_bit(0, &(dev)->tbusy); \
                                                mark_bh(NET_BH); }
#define netif_running(dev)       test_bit(0, &(dev)->start)
#define netif_queue_stopped(dev) test_bit(0, &(dev)->tbusy)
#elif ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,47) )
#define netif_running(dev)       test_bit(LINK_STATE_START, &(dev)->state)
#define netif_queue_stopped(dev) test_bit(LINK_STATE_XOFF,  &(dev)->state)
#endif

/* Softnet changes also affected how SKBs are handled
 * Special calls need to be made now while in an interrupt handler
 */
#if   ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43) )
#define dev_kfree_skb_irq(skb) dev_kfree_skb(skb)
#endif

/*****************************************************************************
 **
 **  General Module / Driver / Kernel API Changes
 **
 *****************************************************************************/

/* New module_init macro added in 2.3.13 - replaces init_module entry point
 * If MODULE is defined, it expands to an init_module definition
 * If the driver is staticly linked to the kernel, it creates the proper 
 * function	pointer for the initialization routine to be called
 * (no more Space.c)
 * module_exit does the same thing for cleanup_module
 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,13) )
#define module_init(fn) int  init_module   (void) { return fn(); }
#define module_exit(fn) void cleanup_module(void) { return fn(); }
#endif

#endif /* E1000_LINUX_KERNEL_COMPAT_H */

